"""
Contains the Datum class
"""

import collections

import numpy as np
from decimal import Decimal


class Datum(collections.namedtuple('Datum', 'label units data comment doi glossary')):
    """Facilitates the storage of quantum chemical results by labeling them with basic metadata.

    Attributes
    ----------
    label : str
        Official label for `data`, often qcvar. May contain spaces.
    units : str
        ASCII, LaTeX-like representation of units, without square brackets.
    data : float or Decimal or or :py:class:`numpy.ndarray`
        Value for `label`.
    comment : str, optional
        Additional notes.
    doi : str, optional
        Literature citation or definition DOI link.
    glossary : str, optional
        Extended description or definition.

    """

    def __new__(cls, label, units, data, comment='', doi=None, glossary=''):
        return super(Datum, cls).__new__(cls, label, units, data, comment, doi, glossary)

    def __str__(self, label=''):
        width = 40
        text = ['-' * width, '{:^{width}}'.format('Datum ' + self.label, width=width)]
        if label:
            text.append('{:^{width}}'.format(label, width=width))
        text.append('-' * width)
        text.append('Data:     {}'.format(self.data))
        text.append('Units:    [{}]'.format(self.units))
        text.append('doi:      {}'.format(self.doi))
        text.append('Comment:  {}'.format(self.comment))
        text.append('Glossary: {}'.format(self.glossary))
        text.append('-' * width)
        return '\n'.join(text)

    def to_dict(self):
        dicary = dict(self._asdict())  # dict, not OrderedDict
        for d in ['doi', 'comment', 'glossary']:
            dicary.pop(d)
        if isinstance(self.data, (np.ndarray, np.number)):
            if self.data.dtype == np.complex:
                dicary['data'] = [dicary['data'].real.tolist(), dicary['data'].imag.tolist()]
            else:
                dicary['data'] = dicary['data'].tolist()
        elif isinstance(self.data, (complex, np.complex)):
            dicary['data'] = [self.data.real, self.data.imag]

        return dicary

    def to_units(self, units=None):
        from .physical_constants import constants

        to_unit = self.units if units is None else units
        factor = constants.conversion_factor(self.units, to_unit)

        if isinstance(self.data, Decimal):
            return factor * float(self.data)
        else:
            return factor * self.data


def print_variables(qcvars):
    """Form a printable representation of qcvariables.

    Parameters
    ----------
    qcvars : dict of Datum
        Group of Datum objects to print.

    Returns
    -------
    str
        Printable string representation of label, data, and unit in Datum-s.

    """
    text = ['\n  Variable Map:', '  ----------------------------------------------------------------------------']

    if len(qcvars) == 0:
        text.append('  (none)')
        return '\n'.join(text)

    largest_key = max(len(k) for k in qcvars) + 2  # for quotation marks
    largest_characteristic = 8
    for k, v in qcvars.items():
        try:
            exp = int(str(v.data).split('E')[1])
        except IndexError:
            pass
        else:
            largest_characteristic = max(exp, largest_characteristic)

    for k, qca in sorted(qcvars.items()):
        #if k != qca.lbl:
        #    raise ValidationError('Huh? {} != {}'.format(k, qca.label))

        if isinstance(qca.data, np.ndarray):
            data = np.array_str(qca.data, max_line_width=120, precision=8, suppress_small=True)
            data = '\n'.join('        ' + ln for ln in data.splitlines())
            text.append("""  {:{keywidth}} => {:{width}} [{}]""".format(
                '"' + k + '"', '', qca.units, keywidth=largest_key, width=largest_characteristic + 14))
            text.append(data)
        elif isinstance(qca.data, Decimal):
            text.append("""  {:{keywidth}} => {:{width}} [{}]""".format(
                '"' + k + '"', qca.data, qca.units, keywidth=largest_key, width=largest_characteristic + 14))
        else:
            text.append("""  {:{keywidth}} => {:{width}.{prec}f} [{}]""".format(
                '"' + k + '"', qca.data, qca.units, keywidth=largest_key, width=largest_characteristic + 14, prec=12))

    text.append('')
    return '\n'.join(text)



import sys

import utils_compare


def _true_false_decorator(compare_fn, *args, **kwargs):
    """Turns `compare_fn` that returns `None` on success and raises
    `_TestComparisonError` on failure into a function that returns
    True/False, suitable for assertions in pytest.

    """

    def true_false_wrapper(*args, **kwargs):
        try:
            compare_fn(*args, **kwargs)
        except utils_compare._TestComparisonError as err:
            return False
        else:
            return True

    return true_false_wrapper


compare_values = _true_false_decorator(utils_compare.compare_values)
compare_strings = _true_false_decorator(utils_compare.compare_strings)
compare_integers = _true_false_decorator(utils_compare.compare_integers)
#compare_matrices = _true_false_decorator(utils_compare.compare_matrices)
#compare_arrays = _true_false_decorator(utils_compare.compare_arrays)
compare_dicts = _true_false_decorator(utils_compare.compare_dicts)
compare_molrecs = _true_false_decorator(utils_compare.compare_molrecs)


def tnm():
    """Returns the name of the calling function, usually name of test case."""

    return sys._getframe().f_back.f_code.co_name



import copy

import numpy as np
import pytest

from utils import *

import qcelemental

_string_prov_stamp = {'creator': 'QCElemental', 'version': '1.0', 'routine': 'qcelemental.molparse.from_string'}
_schema_prov_stamp = {'creator': 'QCElemental', 'version': '1.0', 'routine': 'qcelemental.molparse.from_schema'}

subject14 = """0 3\n--\nHe 0 0 -5\n--\n@He 0 0 5\nunits au"""

schema14_1 = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule" : {
    "geometry": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
    "symbols": ["He", "He"],
    "atomic_numbers": [2, 2],
    "mass_numbers": [4, 4],
    "atom_labels": ['', ''],
    'fragments': [[0], [1]],
    'fragment_charges': [0.0, 0.0],
    'fragment_multiplicities': [3, 1],
    'masses': [4.00260325413, 4.00260325413],
    'name': 'He2',
    'fix_com': False,
    'fix_orientation': False,
    'molecular_charge': 0.0,
    "molecular_multiplicity": 3,
    "real": [True, False]
}}

schema14_psi4 = {
    "geom": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
    "elem": ["He", "He"],
    'elea': [4, 4],
    'elez': [2, 2],
    'fragment_charges': [0.0, 0.0],
    'fragment_multiplicities': [3, 1],
    'mass': [4.00260325413, 4.00260325413],
    'name': 'He2',
    'fix_com': False,
    'fix_orientation': False,
    'molecular_charge': 0.0,
    "molecular_multiplicity": 3,
    'units': 'Bohr',
    'fragment_separators': [1],
    'elbl': ['', ''],
    "real": [True, False]
}


def test_1_14a():
    fullans = copy.deepcopy(schema14_1)
    fullans['molecule']['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype=1)
    assert compare_molrecs(fullans['molecule'], kmol['molecule'], 4, tnm())

    fullans = copy.deepcopy(schema14_psi4)
    fullans['provenance'] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec, 4, tnm())


def test_1_ang_14b():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final['qm'], dtype=1, units='Angstrom')

    assert "QC_JSON_Schema 1 allows only 'Bohr' coordinates" in str(e)


def test_psi4_14c():
    fullans = copy.deepcopy(schema14_psi4)
    fullans['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype='psi4')
    assert compare_molrecs(fullans, kmol, 4, tnm())


def test_dtype_d():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final['qm'], dtype='xkcd927')

    assert "Schema dtype not understood" in str(e)


def test_psi4_nm_14e():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final['qm'], dtype='psi4', units='nm')

    assert "Psi4 Schema psi4 allows only 'Bohr'/'Angstrom' coordinates, not nm" in str(e)


twobohrinang = 2.0 * qcelemental.constants.conversion_factor("bohr", "angstrom")
subject15 = """symmetry cS\nH 0 0 {twobohrinang}\nO 0 0 0\n2H_deut {twobohrinang} 0 0\nno_com\nno_reorient""".format(twobohrinang=twobohrinang)

schema15_1 = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule" : {
    "geometry": [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0],
    "symbols": ["H", "O", "H"],
    "atomic_numbers": [1, 8, 1],
    "mass_numbers": [1, 16, 2],
    "atom_labels": ['', '', '_deut'],
    'fragments': [[0, 1, 2]],
    'fragment_charges': [0.0],
    'fragment_multiplicities': [1],
    'masses': [1.00782503223, 15.99491461957, 2.01410177812],
    'name': 'H2O',
    'comment': 'I has a comment',
    'fix_com': True,
    'fix_orientation': True,
    'fix_symmetry': 'cs',
    'molecular_charge': 0.0,
    "molecular_multiplicity": 1,
    "real": [True, True, True]
}}

schema15_psi4 = {
    "geom": [0.0, 0.0, twobohrinang, 0.0, 0.0, 0.0, twobohrinang, 0.0, 0.0],
    "elem": ["H", "O", "H"],
    "elez": [1, 8, 1],
    "elea": [1, 16, 2],
    "elbl": ['', '', '_deut'],
    'fragment_charges': [0.0],
    'fragment_multiplicities': [1],
    'mass': [1.00782503223, 15.99491461957, 2.01410177812],
    'name': 'H2O',
    'comment': 'I has a comment',
    'fix_com': True,
    'fix_orientation': True,
    'fix_symmetry': 'cs',
    'molecular_charge': 0.0,
    "molecular_multiplicity": 1,
    'units': 'Angstrom',
    'fragment_separators': [],
    'elbl': ['', '', '_deut'],
    "real": [True, True, True]
}


def test_1_15a():
    fullans = copy.deepcopy(schema15_1)
    fullans['molecule']['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final['qm']['comment'] = 'I has a comment'
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype=1)
    assert compare_molrecs(fullans['molecule'], kmol['molecule'], 4, tnm())

    fullans = copy.deepcopy(schema15_psi4)
    fullans['units'] = 'Bohr'
    fullans['geom'] = [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0]
    fullans['provenance'] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec, 4, tnm())


def test_psi4_15c():
    fullans = copy.deepcopy(schema15_psi4)
    fullans['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final['qm']['comment'] = 'I has a comment'
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype='psi4', units='Angstrom')
    assert compare_molrecs(fullans, kmol, 4, tnm())


schema16_1 = {
        'schema_name': 'qc_schema_input',
        'schema_version': 1,
        'molecule': {
            'geometry': [2., 2., 3.],
            'symbols': ['C'],
            'masses': [12.0],
            'atom_labels': [''],
            'atomic_numbers': [6],
            'mass_numbers': [12],
            'real': [True],
            'name': 'C',
            'molecular_charge': 0.,
            'molecular_multiplicity': 1,
            'fragments': [[0]],
            'fragment_charges': [0.],
            'fragment_multiplicities': [1],
            'fix_com': False,
            'fix_orientation': False,
            'provenance': {
                'creator': 'Mystery Program',
                'version': '2018.3',
                'routine': 'molecule builder',
            },
            'connectivity': [(0, 0, 0.0)],
        },
    }

schema16_psi4 = {
            'units': 'Bohr',
            'geom': np.array([2., 2., 3.]),
            'elem': np.array(['C']),
            'mass': np.array([12.0]),
            'elbl': np.array(['']),
            'elez': np.array([6]),
            'elea': np.array([12]),
            'real': np.array([True]),
            'name': 'C',
            'molecular_charge': 0.,
            'molecular_multiplicity': 1,
            'fragment_separators': [],
            'fragment_charges': [0.],
            'fragment_multiplicities': [1],
            'fix_com': False,
            'fix_orientation': False,
            'provenance': {
                'creator': 'Mystery Program',
                'version': '2018.3',
                'routine': 'molecule builder',
            },
            'connectivity': [(0, 0, 0.0)],
    }

def test_froto_16a():
    basic = {
        'schema_name': 'qc_schema_output',
        'schema_version': 1,
        'molecule': {
            'geometry': [2, 2, 3],
            'symbols': ['C'],
            'connectivity': [(0.0, -0.0, 0)],
            'provenance': {
                'creator': 'Mystery Program',
                'version': '2018.3',
                'routine': 'molecule builder',
            },
        },
    }

    fullans = copy.deepcopy(schema16_1)
    fullans['molecule']['provenance'] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.to_schema(qcelemental.molparse.from_schema(basic), dtype=1)
    assert compare_molrecs(fullans['molecule'], roundtrip['molecule'], 4, tnm())


def test_tofro_16b():

    fullans = copy.deepcopy(schema16_psi4)
    fullans['provenance'] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.from_schema(qcelemental.molparse.to_schema(schema16_psi4, dtype=1))
    assert compare_molrecs(fullans, roundtrip, 4, tnm())



#from .np_blockwise import blockwise_expand, blockwise_contract
#from .np_rand3drot import random_rotation_matrix
#from .mpl import plot_coord
from .misc import distance_matrix, update_with_error, standardize_efp_angles_units, filter_comments, unnp
from .internal import provenance_stamp
from .itertools import unique_everseen



from pydantic import BaseModel
from enum import Enum
from typing import Any
import numpy as np

ndarray_encoder = {np.ndarray: lambda v: v.flatten().tolist()}


class Provenance(BaseModel):
    creator: str
    version: str = None
    routine: str = None

    class Config:
        allow_extra = True


class Model(BaseModel):
    method: str
    basis: str = None
    # basis_spec: BasisSpec = None  # This should be exclusive with basis, but for now will be omitted

    class Config:
        allow_mutation = False
        allow_extra = True


class DriverEnum(str, Enum):
    energy = 'energy'
    gradient = 'gradient'
    hessian = 'hessian'


class ComputeError(BaseModel):
    """The type of error message raised"""
    error_type: str  # Error enumeration not yet strict
    error_message: str

    class Config:
        allow_extra = False


class FailedOperation(BaseModel):
    id: str = None
    input_data: Any = None
    success: bool = False
    error: ComputeError

    class Config:
        allow_extra = True
        allow_mutation = False
        json_encoders = {
            **ndarray_encoder
        }


qcschema_input_default = "qcschema_input"
qcschema_output_default = "qcschema_output"
qcschema_optimization_input_default = "qcschema_optimization_input"
qcschema_optimization_output_default = "qcschema_optimization_output"



"""
Molecule Object Model
"""

import os
import json
import hashlib
import collections
import numpy as np
from pydantic import BaseModel, validator
from typing import List, Tuple
from ..physical_constants import constants
from ..periodic_table import periodictable
from .common_models import Provenance, ndarray_encoder
from ..util import provenance_stamp

from ..molparse import from_string, from_arrays, to_schema

# Rounding quantities for hashing
GEOMETRY_NOISE = 8
MASS_NOISE = 6
CHARGE_NOISE = 4


def float_prep(array, around):
    """
    Rounds floats to a common value and build positive zero's to prevent hash conflicts.
    """
    if isinstance(array, (list, np.ndarray)):
        # Round array
        array = np.around(array, around)
        # Flip zeros
        array[np.abs(array) < 5**(-(around + 1))] = 0

    elif isinstance(array, (float, int)):
        array = round(array, around)
        if array == -0.0:
            array = 0.0
    else:
        raise TypeError("Type '{}' not recognized".format(type(array).__name__))

    return array


class NPArray(np.ndarray):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        try:
            v = np.array(v, dtype=np.double)
        except:
            raise RuntimeError("Could not cast {} to NumPy Array!".format(v))
        return v


class Identifiers(BaseModel):
    """Canonical chemical identifiers"""

    # molecule_hash: str = None
    # molecular_formula: str = None
    # smiles: str = None
    # inchi: str = None
    # inchikey: str = None
    # canonical_explicit_hydrogen_smiles: str = None
    # canonical_isomeric_explicit_hydrogen_mapped_smiles: str = None
    # canonical_isomeric_explicit_hydrogen_smiles: str = None
    # canonical_isomeric_smiles: str = None
    # canonical_smiles: str = None

    class Config:
        allow_extra = True


# Cleanup un-initialized variables


class Molecule(BaseModel):

    # Required data
    symbols: List[str]
    # geometry: List[float]
    geometry: NPArray

    # Molecule data
    name: str = ""
    identifiers: Identifiers = None
    comment: str = None
    molecular_charge: float = 0.0
    molecular_multiplicity: int = 1

    # Atom data
    masses: List[float] = None
    real: List[bool] = None
    atom_labels: List[str] = None
    atomic_numbers: List[int] = None
    mass_numbers: List[int] = None

    # Fragment and connection data
    connectivity: List[Tuple[int, int, float]] = []
    fragments: List[List[int]] = None
    fragment_charges: List[float] = None
    fragment_multiplicities: List[int] = None

    # Orientation
    fix_com: bool = False
    fix_orientation: bool = False

    # Extra
    provenance: Provenance = provenance_stamp(__name__)
    id: str = None

    class Config:
        json_encoders = {**ndarray_encoder}
        allow_mutation = False
        ignore_extra = False

    # Internal values as a mutable object we can manipulate
    class _Internals:
        provided_fields = set()

    def __init__(self, orient=False, **kwargs):
        super().__init__(**kwargs)
        self.Config.allow_mutation = True  # Set this to set some immutability config
        self._Internals.provided_fields = set(kwargs.keys())
        self.symbols = [s.title() for s in self.symbols]  # Title case

        if self.masses is None:  # Setup masses before fixing the orientation
            self.masses = [periodictable.to_mass(x) for x in self.symbols]

        if self.real is None:
            self.real = [True for _ in self.symbols]

        if orient:
            self.geometry = float_prep(self._orient_molecule_internal(), GEOMETRY_NOISE)
        else:
            self.geometry = float_prep(self.geometry, GEOMETRY_NOISE)

        # Cleanup un-initialized variables  (more complex than Pydantic Validators allow)
        if not self.fragments:
            natoms = self.geometry.shape[0]
            self.fragments = [list(range(natoms))]
            self.fragment_charges = [self.molecular_charge]
            self.fragment_multiplicities = [self.molecular_multiplicity]
        else:
            if not self.fragment_charges:
                if np.isclose(self.molecular_charge, 0.0):
                    self.fragment_charges = [0 for _ in self.fragments]
                else:
                    raise KeyError("Fragments passed in, but not fragment charges for a charged molecule.")

            if not self.fragment_multiplicities:
                if self.molecular_multiplicity == 1:
                    self.fragment_multiplicities = [1 for _ in self.fragments]
                else:
                    raise KeyError("Fragments passed in, but not fragment multiplicities for a non-singlet molecule.")

        self.Config.allow_mutation = False  # Reset mutation

    @validator('geometry')
    def must_be_3n(cls, v, values, **kwargs):
        n = len(values['symbols'])
        try:
            v = v.reshape(n, 3)
            if v.shape != (n, 3):
                raise ValueError()
        except (ValueError, AttributeError):
            raise ValueError("Geometry must be castable to shape (N,3)!")
        return v

    @validator('masses', 'real', whole=True)
    def must_be_n(cls, v, values, **kwargs):
        n = len(values['symbols'])
        if len(v) != n:
            raise ValueError("Masses and Real must be same number of entries as Symbols")
        return v

    @validator('real', whole=True)
    def populate_real(cls, v, values, **kwargs):
        # Can't use geometry here since its already been validated and not in values
        n = len(values['symbols'])
        if len(v) == 0:
            v = [True for _ in range(n)]
        return v

    @validator('fragment_charges', 'fragment_multiplicities', whole=True)
    def must_be_n_frag(cls, v, values, **kwargs):
        if 'fragments' in values:
            n = len(values['fragments'])
            if len(v) != n:
                raise ValueError("Fragment Charges and Fragment Multiplicities"
                                 " must be same number of entries as Fragments")
        else:
            raise ValueError("Cannot have Fragment Charges or Fragment Multiplicities " "without Fragments")
        return v

    @validator('connectivity')
    def min_zero(cls, v):
        if v < 0:
            raise ValueError("Connectivity entries must be greater than 0")
        return v

    @property
    def hash_fields(self):
        return [
            "symbols", "masses", "molecular_charge", "molecular_multiplicity", "real", "geometry", "fragments",
            "fragment_charges", "fragment_multiplicities", "connectivity"
        ]

    def dict(self, *args, **kwargs):
        if ("include" not in kwargs) or (("include" in kwargs) and (kwargs["include"] is None)):
            kwargs["include"] = self._Internals.provided_fields
            if self.id is None:
                kwargs["include"] -= {"id"}

        return super().dict(*args, **kwargs)

    def json(self, *args, **kwargs):
        as_dict = kwargs.pop("as_dict", False)

        ret = super().json(*args, **kwargs)
        if as_dict:
            return json.loads(ret)
        else:
            return ret

### Non-Pydantic API functions

    def orient_molecule(self):
        """
        Centers the molecule and orients via inertia tensor before returning a new Molecule
        """
        return Molecule(orient=True, **self.dict())

    def compare(self, other, bench=None):
        """
        Checks if two molecules are identical. This is a molecular identity defined
        by scientific terms, and not programing terms, so its less rigorous than
        a programmatic equality or a memory equivalent `is`.
        """

        if isinstance(other, dict):
            other = Molecule(orient=False, **other)
        elif isinstance(other, Molecule):
            pass
        else:
            raise TypeError("Comparison molecule not understood of type '{}'.".format(type(other)))

        if bench is None:
            bench = self

        match = True
        match &= bench.symbols == other.symbols
        match &= np.allclose(bench.masses, other.masses, atol=MASS_NOISE)
        match &= np.equal(bench.real, other.real).all()
        match &= np.equal(bench.fragments, other.fragments).all()
        match &= np.allclose(bench.fragment_charges, other.fragment_charges, atol=CHARGE_NOISE)
        match &= np.equal(bench.fragment_multiplicities, other.fragment_multiplicities).all()

        match &= np.allclose(bench.molecular_charge, other.molecular_charge, atol=CHARGE_NOISE)
        match &= np.equal(bench.molecular_multiplicity, other.molecular_multiplicity).all()
        match &= np.allclose(bench.geometry, other.geometry, atol=GEOMETRY_NOISE)
        return match

    def pretty_print(self):
        """Print the molecule in Angstroms. Same as :py:func:`print_out` only always in Angstroms.
        (method name in libmints is print_in_angstrom)

        """
        text = ""

        text += """    Geometry (in {0:s}), charge = {1:.1f}, multiplicity = {2:d}:\n\n""".format(
            'Angstrom', self.molecular_charge, self.molecular_multiplicity)
        text += """       Center              X                  Y                   Z       \n"""
        text += """    ------------   -----------------  -----------------  -----------------\n"""

        for i in range(len(self.geometry)):
            text += """    {0:8s}{1:4s} """.format(self.symbols[i], "" if self.real[i] else "(Gh)")
            for j in range(3):
                text += """  {0:17.12f}""".format(
                    self.geometry[i][j] * constants.conversion_factor("bohr", "angstroms"))
            text += "\n"
        text += "\n"

        return text

    def get_fragment(self, real, ghost=None, orient=False):
        """
        A list of real and ghost fragments:
        """

        if isinstance(real, int):
            real = [real]

        if isinstance(ghost, int):
            ghost = [ghost]
        elif ghost is None:
            ghost = []

        constructor_dict = {}

        ret_name = self.name + " (" + str(real) + "," + str(ghost) + ")"
        constructor_dict["name"] = ret_name
        # ret = Molecule(None, name=ret_name)

        if len(set(real) & set(ghost)):
            raise TypeError("Molecule:get_fragment: real and ghost sets are overlapping! ({0}, {1}).".format(
                str(real), str(ghost)))

        geom_blocks = []
        symbols = []
        masses = []
        real_atoms = []
        fragments = []
        fragment_charges = []
        fragment_multiplicities = []

        # Loop through the real blocks
        frag_start = 0
        for frag in real:
            frag_size = len(self.fragments[frag])
            geom_blocks.append(self.geometry[self.fragments[frag]])

            for idx in self.fragments[frag]:
                symbols.append(self.symbols[idx])
                real_atoms.append(True)
                masses.append(self.masses[idx])

            fragments.append(list(range(frag_start, frag_start + frag_size)))
            frag_start += frag_size

            fragment_charges.append(float(self.fragment_charges[frag]))
            fragment_multiplicities.append(self.fragment_multiplicities[frag])

        # Set charge and multiplicity
        constructor_dict["molecular_charge"] = sum(fragment_charges)
        constructor_dict["molecular_multiplicity"] = sum(x - 1 for x in fragment_multiplicities) + 1

        # Loop through the ghost blocks
        for frag in ghost:
            frag_size = len(self.fragments[frag])
            geom_blocks.append(self.geometry[self.fragments[frag]])

            for idx in self.fragments[frag]:
                symbols.append(self.symbols[idx])
                real_atoms.append(False)
                masses.append(self.masses[idx])

            fragments.append(list(range(frag_start, frag_start + frag_size)))
            frag_start += frag_size

            fragment_charges.append(self.fragment_charges[frag])
            fragment_multiplicities.append(self.fragment_multiplicities[frag])

        constructor_dict["fragments"] = fragments
        constructor_dict["symbols"] = symbols
        constructor_dict["geometry"] = np.vstack(geom_blocks)
        constructor_dict["real"] = real_atoms
        constructor_dict["masses"] = masses

        return Molecule(orient=orient, **constructor_dict)

    def to_string(self, dtype="psi4"):
        """Returns a string that can be used by a variety of programs.

        Unclear if this will be removed or renamed to "to_psi4_string" in the future
        """
        if dtype == "psi4":
            return self._to_psi4_string()
        else:
            raise KeyError("Molecule:to_string: dtype of '{}' not recognized.".format(dtype))

    def get_hash(self):
        """
        Returns the hash of the molecule.
        """

        m = hashlib.sha1()
        concat = ""

        tmp_dict = super().dict()

        np.set_printoptions(precision=16)
        for field in self.hash_fields:
            data = tmp_dict[field]
            if field == "geometry":
                tmp_dict[field] = float_prep(data, GEOMETRY_NOISE).ravel().tolist()
            elif field == "fragment_charges":
                tmp_dict[field] = float_prep(data, CHARGE_NOISE).tolist()
            elif field == "molecular_charge":
                tmp_dict[field] = float_prep(data, CHARGE_NOISE)
            elif field == "masses":
                tmp_dict[field] = float_prep(data, MASS_NOISE).tolist()

            concat += json.dumps(tmp_dict[field])  # This should only be operating on Python types now

        m.update(concat.encode("utf-8"))
        return m.hexdigest()

    def get_molecular_formula(self):
        """
        Returns the molecular formula for a molecule. Atom symbols are sorted from
        A-Z.

        Examples
        --------

        >>> methane = qcelemental.models.Molecule('''
        ... H      0.5288      0.1610      0.9359
        ... C      0.0000      0.0000      0.0000
        ... H      0.2051      0.8240     -0.6786
        ... H      0.3345     -0.9314     -0.4496
        ... H     -1.0685     -0.0537      0.1921
        ... ''')

        >>> methane.get_molecular_formula()
        CH4

        >>> hcl = qcelemental.models.Molecule('''
        ... H      0.0000      0.0000      0.0000
        ... Cl     0.0000      0.0000      1.2000
        ... ''')

        >>> hcl.get_molecular_formula()
        ClH

        """
        count = collections.Counter(x.title() for x in self.symbols)

        ret = []
        for k in sorted(count.keys()):
            c = count[k]
            ret.append(k)
            if c > 1:
                ret.append(str(c))

        return "".join(ret)

    ### Constructors

    @classmethod
    def from_data(cls, data, dtype=None, *, orient=False, **kwargs):
        """
        Constructs a molecule object from a data structure.

        Parameters
        ----------
        data : Object
            Data to construct Molecule from. This is likely what would be loaded from a file
        dtype : {"string", "numpy", "json", "dict"}
            The type of data to interpret,
            no attempt to infer what type of data is done here, it must be provided
        orient: bool, optional
            Orientates the molecule to a standard frame or not.
        kwargs
            Any additional keywords to pass to the constructor

        Returns
        -------
        Molecule
            A constructed molecule class.
        """
        if dtype is None:
            if isinstance(data, str):
                dtype = "string"
            elif isinstance(dat, np.array):
                dtype = "numpy"
            elif isinstance(data, dict):
                dtype = "dict"
            else:
                raise TypeError("Input type not understood, please supply the 'dtype' kwarg.")

        if dtype in ["string", "psi4", "psi4+", "xyz", "xyz+"]:
            input_dict = to_schema(from_string(data)["qm"], dtype=1)["molecule"]
        elif dtype == "numpy":
            data = np.asarray(data)
            data = {
                "geom": data[:, 1:],
                "elez": data[:, 0],
                "units": kwargs.pop("units", "Angstrom"),
                "fragment_separators": kwargs.pop("frags", [])
            }
            input_dict = to_schema(from_arrays(**data), dtype=1)["molecule"]
        elif dtype == "json":
            input_dict = json.loads(data)
        elif dtype == "dict":
            input_dict = data
        else:
            raise KeyError("Dtype not understood '{}'.".format(dtype))

        return cls(orient=orient, **input_dict)

    @classmethod
    def from_file(cls, filename, dtype=None, orient=False, **kwargs):
        """
        Constructs a molecule object from a file.

        Parameters
        ----------
        filename : str
            The filename to build
        dtype : {None, "psi4", "numpy", "json"}, optional
            The type of file to interpret.
        orient: bool, optional
            Orientates the molecule to a standard frame or not.
        kwargs
            Any additional keywords to pass to the constructor

        Returns
        -------
        Molecule
            A constructed molecule class.
        """

        ext = os.path.splitext(filename)[1]

        if dtype is None:
            if ext in [".psimol"]:
                dtype = "psi4"
            elif ext in [".npy"]:
                dtype = "numpy"
            elif ext in [".json"]:
                dtype = "json"
            else:
                raise KeyError("No dtype provided and ext '{}' not understood.".format(ext))
            # print("Inferring data type to be {} from file extension".format(dtype))

        if dtype == "psi4":
            with open(filename, "r") as infile:
                data = infile.read()
        elif dtype == "numpy":
            data = np.load(filename)
        elif dtype == "json":
            with open(filename, "r") as infile:
                data = json.load(infile)
            dtype = "dict"
        else:
            raise KeyError("Dtype not understood '{}'.".format(dtype))

        return cls.from_data(data, dtype, orient=orient, **kwargs)

    ### Non-Pydantic internal functions

    def _orient_molecule_internal(self):
        """
        Centers the molecule and orients via inertia tensor before returning a new set of the
        molecule geometry
        """

        new_geometry = self.geometry.copy()  # Ensure we get a copy
        # Get the mass as an array
        # Masses are needed for orientation
        np_mass = np.array(self.masses)

        # Center on Mass
        new_geometry -= np.average(new_geometry, axis=0, weights=np_mass)

        # Rotate into inertial frame
        tensor = self._inertial_tensor(new_geometry, np_mass)
        evals, evecs = np.linalg.eigh(tensor)

        new_geometry = np.dot(new_geometry, evecs)

        # Phases? Lets do the simplest thing and ensure the first atom in each column
        # that is not on a plane is positve

        phase_check = [False, False, False]

        geom_noise = 10**(-GEOMETRY_NOISE)
        for num in range(new_geometry.shape[0]):

            for x in range(3):
                if phase_check[x]:
                    continue

                val = new_geometry[num, x]

                if abs(val) < geom_noise:
                    continue

                phase_check[x] = True

                if val < 0:
                    new_geometry[:, x] *= -1

            if sum(phase_check) == 3:
                break
        return new_geometry

    def __str__(self):
        return self.pretty_print()

    @staticmethod
    def _inertial_tensor(geom, weight):
        """
        Compute the moment inertia tensor for a given geometry.
        """
        # Build inertia tensor
        tensor = np.zeros((3, 3))

        # Diagonal
        tensor[0][0] = np.sum(weight * (geom[:, 1]**2.0 + geom[:, 2]**2.0))
        tensor[1][1] = np.sum(weight * (geom[:, 0]**2.0 + geom[:, 2]**2.0))
        tensor[2][2] = np.sum(weight * (geom[:, 0]**2.0 + geom[:, 1]**2.0))

        # I(alpha, beta)
        # Off diagonal
        tensor[1][0] = tensor[0][1] = -1.0 * np.sum(weight * geom[:, 0] * geom[:, 1])
        tensor[2][0] = tensor[0][2] = -1.0 * np.sum(weight * geom[:, 0] * geom[:, 2])
        tensor[2][1] = tensor[1][2] = -1.0 * np.sum(weight * geom[:, 1] * geom[:, 2])
        return tensor

    def _to_psi4_string(self):
        """Regenerates a input file molecule specification string from the
        current state of the Molecule. Contains geometry info,
        fragmentation, charges and multiplicities, and any frame
        restriction.
        """
        text = "\n"

        # append atoms and coordinates and fragment separators with charge and multiplicity
        for num, frag in enumerate(self.fragments):
            divider = "    --"
            if num == 0:
                divider = ""

            if any(self.real[at] for at in frag):
                text += "{0:s}    \n    {1:d} {2:d}\n".format(divider,
                                                              int(self.fragment_charges[num]),
                                                              self.fragment_multiplicities[num])

            for at in frag:
                if self.real[at]:
                    text += "    {0:<8s}".format(str(self.symbols[at]))
                else:
                    text += "    {0:<8s}".format("Gh(" + self.symbols[at] + ")")
                text += "    {0: 14.10f} {1: 14.10f} {2: 14.10f}\n".format(*tuple(self.geometry[at]))
        text += "\n"

        # append units and any other non-default molecule keywords
        text += "    units bohr\n"
        text += "    no_com\n"
        text += "    no_reorient\n"

        return text



import numpy as np
import pytest
from decimal import Decimal

from utils import *

import qcelemental as qcel


@pytest.fixture
def dataset():
    datums = {
        'decimal': qcel.Datum('a label', 'mdyn/angstrom', Decimal('4.4'), 'force constant', '10.1000/182'),
        'ndarray': qcel.Datum('an array', 'cm^-1',
                              np.arange(4, dtype=np.float) * 4 / 3, 'freqs'),
        'float': qcel.Datum('a float', 'kg', 4.4, doi='10.1000/182'),
    }

    return datums


def test_creation(dataset):
    datum1 = dataset['decimal']

    assert datum1.label == 'a label'
    assert datum1.units == 'mdyn/angstrom'
    assert datum1.data == Decimal('4.4')


@pytest.mark.parametrize("inp,expected", [
    (('decimal', None), 4.4),
    (('decimal', 'N/m'), 440),
    (('decimal', 'hartree/bohr/bohr'), 0.282614141011),
    (('ndarray', '1/m'), np.arange(4, dtype=np.float) * 400 / 3),
])
def test_units(dataset, inp, expected):
    assert dataset[inp[0]].to_units(inp[1]) == pytest.approx(expected, 1.e-9)


def test_printing(dataset):
    datum1 = dataset['decimal']
    str1 = """----------------------------------------
             Datum a label              
                 Pytest                 
----------------------------------------
Data:     4.4
Units:    [mdyn/angstrom]
doi:      10.1000/182
Comment:  force constant
Glossary: 
----------------------------------------"""

    assert str1 == datum1.__str__(label='Pytest')


def test_mass_printing_blank():
    pvnone = """
  Variable Map:
  ----------------------------------------------------------------------------
  (none)"""

    assert pvnone == qcel.datum.print_variables({})


def test_mass_printing(dataset):
    str1 = """
  Variable Map:
  ----------------------------------------------------------------------------
  "decimal" =>                    4.4 [mdyn/angstrom]
  "float"   =>         4.400000000000 [kg]
  "ndarray" =>                        [cm^-1]
        [0.         1.33333333 2.66666667 4.        ]
"""

    assert str1 == qcel.datum.print_variables(dataset)


def test_to_dict(dataset):
    listans = [i * 4 / 3 for i in range(4)]
    ans = {'label': 'an array', 'units': 'cm^-1', 'data': listans}

    dicary = dataset['ndarray'].to_dict()
    assert compare_dicts(ans, dicary, 9, tnm())


def test_complex_scalar():
    datum1 = qcel.Datum('complex scalar', '', complex(1, 2))
    ans = {'label': 'complex scalar', 'units': '', 'data': [1.0, 2.0]}

    assert datum1.label == 'complex scalar'
    assert datum1.units == ''
    assert datum1.data.real == 1
    assert datum1.data.imag == 2

    dicary = datum1.to_dict()
    assert compare_dicts(ans, dicary, 9, tnm())


def test_complex_array():
    datum1 = qcel.Datum('complex array', '', np.arange(3, dtype=np.complex_) + 1j)
    ans = {'label': 'complex array', 'units': '', 'data': [[0.0, 1.0, 2.0], [1.0, 1.0, 1.0]]}

    dicary = datum1.to_dict()
    assert compare_dicts(ans, dicary, 9, tnm())



"""
Units handlers
"""

try:
    import pint
except ImportError:
    raise ImportError("""Python module pint not found. Solve by installing it: `conda install pint -c conda-forge` or `pip install pint`""")

from .context import constants, PhysicalConstantsContext



"""
Main init for QCElemental
"""

from .datum import Datum
from .exceptions import (NotAnElementError, ValidationError, MoleculeFormatError, ChoicesError, DataUnavailableError)
from . import molparse
from . import models

# Handle singletons, not their classes or modules
from .periodic_table import periodictable
from .physical_constants import constants, PhysicalConstantsContext
from .covalent_radii import covalentradii, CovalentRadii
del periodic_table
del physical_constants
del covalent_radii

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



from enum import Enum
from pydantic import BaseModel, constr, validator
from typing import List, Union, Dict, Any
from .molecule import Molecule
from .common_models import (Provenance, Model, DriverEnum, ComputeError, qcschema_input_default,
                            qcschema_output_default, ndarray_encoder)
from ..util import provenance_stamp


class Properties(BaseModel):
    scf_one_electron_energy: float = None
    scf_two_electron_energy: float = None
    nuclear_repulsion_energy: float = None
    scf_vv10_energy: float = None
    scf_xc_energy: float = None
    scf_dispersion_correction_energy: float = None
    scf_dipole_moment: List[float] = None
    scf_total_energy: float = None
    scf_iterations: int = None
    mp2_same_spin_correlation_energy: float = None
    mp2_opposite_spin_correlation_energy: float = None
    mp2_singles_energy: float = None
    mp2_doubles_energy: float = None
    mp2_total_correlation_energy: float = None
    mp2_total_energy: float = None
    calcinfo_nbasis: int = None
    calcinfo_nmo: int = None
    calcinfo_nalpha: int = None
    calcinfo_nbeta: int = None
    calcinfo_natom: int = None
    return_energy: float = None

    class Config:
        allow_extra = True  # Not yet fully validated, but will accept extra for now


class ErrorEnum(str, Enum):
    convergence_error = "convergence_error"
    file_error = "file_error"
    memory_error = "memory_error"


### Primary models


class ResultInput(BaseModel):
    """The MolSSI Quantum Chemistry Schema"""
    id: str = None
    molecule: Molecule
    driver: DriverEnum
    model: Model
    schema_name: constr(strip_whitespace=True, regex=qcschema_input_default) = qcschema_input_default
    schema_version: int = 1
    keywords: dict = {}
    provenance: Provenance = provenance_stamp(__name__)

    class Config:
        allow_mutation = False
        allow_extra = True
        json_encoders = {**ndarray_encoder}

    def dict(self, *args, **kwargs):
        if self.id is None:
            excl = kwargs.setdefault("exclude", [])
            if isinstance(excl, list):
                excl.append("id")
            elif isinstance(excl, set):
                excl |= {"id"}

        return super().dict(*args, **kwargs)


class Result(ResultInput):
    schema_name: constr(strip_whitespace=True, regex=qcschema_output_default) = qcschema_output_default
    properties: Properties = Properties()
    success: bool
    error: ComputeError = None
    return_result: Union[float, List[float], Dict[str, Any]]

    class Config(ResultInput.Config):
        # Will carry other properties
        pass

    @validator("schema_name", pre=True)
    def input_to_output(cls, v):
        """If qcschema_input is passed in, cast it to output, otherwise no"""
        if v.lower().strip() in [qcschema_input_default, qcschema_output_default]:
            return qcschema_output_default
        raise ValueError("Only {0} or {1} is allowed for schema_name, "
                         "which will be converted to {0}".format(qcschema_output_default, qcschema_input_default))



import re
import math

import numpy as np

from ..physical_constants import constants


def distance_matrix(a, b):
    """Euclidean distance matrix between rows of arrays `a` and `b`. Equivalent to
    `scipy.spatial.distance.cdist(a, b, 'euclidean')`. Returns a.shape[0] x b.shape[0] array.

    """
    assert a.shape[1] == b.shape[1], """Inner dimensions do not match"""
    distm = np.zeros([a.shape[0], b.shape[0]])
    for i in range(a.shape[0]):
        distm[i] = np.linalg.norm(a[i] - b, axis=1)
    return distm


def update_with_error(a, b, path=None):
    """Merges `b` into `a` like dict.update; however, raises KeyError if values of a
    key shared by `a` and `b` conflict.

    Adapted from: https://stackoverflow.com/a/7205107

    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                update_with_error(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            elif a[key] is None:
                a[key] = b[key]
            elif (isinstance(a[key], (list, tuple)) and
                  not isinstance(a[key], str) and
                  isinstance(b[key], (list, tuple)) and
                  not isinstance(b[key], str) and
                  len(a[key]) == len(b[key]) and
                  all((av is None or av == bv) for av, bv in zip(a[key], b[key]))):  # yapf: disable
                a[key] = b[key]
            else:
                raise KeyError('Conflict at {}: {} vs. {}'.format('.'.join(path + [str(key)]), a[key], b[key]))
        else:
            a[key] = b[key]
    return a


def standardize_efp_angles_units(units, geom_hints):
    """Applies to the pre-validated xyzabc or points hints in `geom_hints`
    the libefp default (1) units of [a0] and (2) radian angle range of
    (-pi, pi]. The latter is handy since this is how libefp returns hints

    """

    def radrge(radang):
        """Adjust `radang` by 2pi into (-pi, pi] range."""
        if radang > math.pi:
            return radang - 2 * math.pi
        elif radang <= -math.pi:
            return radang + 2 * math.pi
        else:
            return radang

    if units == 'Angstrom':
        iutau = 1. / constants.bohr2angstroms
    else:
        iutau = 1.

    hints = []
    for hint in geom_hints:
        if len(hint) == 6:
            x, y, z = [i * iutau for i in hint[:3]]
            a, b, c = [radrge(i) for i in hint[3:]]
            hints.append([x, y, z, a, b, c])
        if len(hint) == 9:
            points = [i * iutau for i in hint]
            hints.append(points)

    return hints


def filter_comments(string):
    """Remove from `string` any Python-style comments ('#' to end of line)."""

    comment = re.compile(r'(^|[^\\])#.*')
    string = re.sub(comment, '', string)
    return string


def unnp(dicary, flat=False, _path=None):
    """Return `dicary` with any ndarray values replaced by lists.

    Parameters
    ----------
    dicary: dict
        Dictionary where any internal iterables are dict or list.
    flat : bool, optional
        Whether the returned lists are flat or nested.

    Returns
    -------
    dict
        Input with any ndarray values replaced by lists.

    """
    if _path is None:
        _path = []

    ndicary = {}
    for k, v in dicary.items():
        if isinstance(v, dict):
            ndicary[k] = unnp(v, flat, _path + [str(k)])
        elif isinstance(v, list):
            # relying on Py3.6+ ordered dict here
            fakedict = {kk: vv for kk, vv in enumerate(v)}
            tolisted = unnp(fakedict, flat, _path + [str(k)])
            ndicary[k] = list(tolisted.values())
        else:
            try:
                v.shape
            except AttributeError:
                ndicary[k] = v
            else:
                if flat:
                    ndicary[k] = v.ravel().tolist()
                else:
                    ndicary[k] = v.tolist()
    return ndicary



from pydantic import BaseModel, constr
from .common_models import (Model, DriverEnum, ComputeError, qcschema_input_default,
                            qcschema_optimization_input_default, qcschema_optimization_output_default, ndarray_encoder)
from .molecule import Molecule
from .results import Result
from typing import Any, Dict, List


class InputSpecification(BaseModel):
    driver: DriverEnum
    schema_name: constr(strip_whitespace=True, regex=qcschema_input_default) = qcschema_input_default
    schema_version: int = 1
    model: Model
    keywords: Dict[str, Any] = {}

    class Config:
        allow_extra = True
        allow_mutation = False


class OptimizationInput(BaseModel):
    schema_name: constr(
        strip_whitespace=True, regex=qcschema_optimization_input_default) = qcschema_optimization_input_default
    schema_version: int = 1
    keywords: Dict[str, Any] = {}
    input_specification: InputSpecification
    initial_molecule: Molecule

    class Config:
        allow_extra = True
        allow_mutation = False
        json_encoders = {**ndarray_encoder}


class Optimization(OptimizationInput):
    id: str = None
    schema_name: constr(
        strip_whitespace=True, regex=qcschema_optimization_output_default) = qcschema_optimization_output_default
    final_molecule: Molecule
    trajectory: List[Result] = None
    energies: List[float] = None
    success: bool
    error: ComputeError = None

    def dict(self, *args, **kwargs):
        if self.id is None:
            excl = kwargs.setdefault("exclude", [])
            if isinstance(excl, list):
                excl.append("id")
            elif isinstance(excl, set):
                excl |= {"id"}

        return super().dict(*args, **kwargs)



import sys
import copy
import math
import pprint
import collections


class _TestComparisonError(Exception):
    """Error when element or nuclide can't be identified."""

    def __init__(self, msg):
        self.message = '\nTestComparisonError: {}\n\n'.format(msg)


def _success(label):
    """Function to print a '*label*...PASSED' line to screen."""

    print('\t{0:.<66}PASSED'.format(label))
    sys.stdout.flush()


def compare_values(expected, computed, digits, label, passnone=False, verbose=1):
    """Function to compare two values to `digits` precision. Prints :py:func:`_success`.

    Raises
    ------
    _TestComparisonError
        If `computed` differs from `expected` by more than number of `digits`.
        (or by `digits` itself when <= 1 e.g. digits=0.04).

    """
    if passnone:
        if expected is None and computed is None:
            _success(label)
            return

    if digits > 1:
        thresh = 10**-digits
        message = """\t{}: computed value ({:.{digits1}f}) does not match ({:.{digits1}f}) to {digits} digits.""".format(
            label, computed, expected, digits1=int(digits) + 1, digits=digits)
    else:
        thresh = digits
        message = ("\t%s: computed value (%f) does not match (%f) to %f digits." % (label, computed, expected, digits))
    if abs(expected - computed) > thresh:
        raise _TestComparisonError(message)
    if math.isnan(computed):
        message += "\tprobably because the computed value is nan."
        raise _TestComparisonError(message)
    if verbose >= 1:
        _success(label)


def compare_integers(expected, computed, label, verbose=1):
    """Function to compare two integers or strings.

    Raises
    ------
    _TestComparisonError
        If `computed` doesn't exactly match `expected`.

    """
    if expected != computed:
        message = "\t{}: computed value ({}) does not match ({}).".format(label, computed, expected)
        raise _TestComparisonError(message)
    if verbose >= 1:
        _success(label)


compare_strings = compare_integers

#def compare_matrices(expected, computed, digits, label, verbose=1):
#    """Function to compare two matrices. Prints :py:func:`util.success`
#    when elements of matrix *computed* match elements of matrix *expected* to
#    number of *digits*. Performs a system exit on failure to match symmetry
#    structure, dimensions, or element values. Used in input files in the test suite.
#
#    """
#    rows = len(expected)
#    cols = len(expected[0])
#    failed = 0
#    for row in range(rows):
#        for col in range(cols):
#            if abs(expected[row][col] - computed[row][col]) > 10 ** (-digits):
#                print("\t%s: computed value (%s) does not match (%s)." % (label, expected[row][col], computed[row][col]))
#                failed = 1
#                break
#
#    if failed:
#        print("The Failed Test Matrices\n")
#        show(computed)
#        print('\n')
#        show(expected)
#        raise _TestComparisonError('compare_matrices failed')
#    if verbose >= 1:
#        _success(label)


def compare_dicts(expected, computed, tol, label, forgive=None, verbose=1):
    """Compares dictionaries `computed` to `expected` using DeepDiff Float
    comparisons made to `tol` significant decimal places. Note that a clean
    DeepDiff returns {}, which evaluates to False, hence the compare_integers.
    Keys in `forgive` may change between `expected` and `computed` without
    triggering failure.

    """
    try:
        import deepdiff
    except ImportError:
        raise ImportError("""Install deepdiff. `conda install deepdiff -c conda-forge` or `pip install deepdiff`""")

    if forgive is None:
        forgive = []
    forgiven = collections.defaultdict(dict)

    ans = deepdiff.DeepDiff(expected, computed, significant_digits=tol, verbose_level=2)

    for category in list(ans):
        for key in list(ans[category]):
            for fg in forgive:
                fgsig = "root['" + fg + "']"
                if key.startswith(fgsig):
                    forgiven[category][key] = ans[category].pop(key)
        if not ans[category]:
            del ans[category]

    clean = not bool(ans)
    if not clean:
        pprint.pprint(ans)
    if verbose >= 2:
        pprint.pprint(forgiven)
    return compare_integers(True, clean, label, verbose=verbose)


def compare_molrecs(expected, computed, tol, label, forgive=None, verbose=1, relative_geoms='exact'):
    """Function to compare Molecule dictionaries. Prints
#    :py:func:`util.success` when elements of `computed` match elements of
#    `expected` to `tol` number of digits (for float arrays).

    """
    thresh = 10**-tol if tol >= 1 else tol

    # Need to manipulate the dictionaries a bit, so hold values
    xptd = copy.deepcopy(expected)
    cptd = copy.deepcopy(computed)

    def massage_dicts(dicary):
        # deepdiff can't cope with np.int type
        #   https://github.com/seperman/deepdiff/issues/97
        if 'elez' in dicary:
            dicary['elez'] = [int(z) for z in dicary['elez']]
        if 'elea' in dicary:
            dicary['elea'] = [int(a) for a in dicary['elea']]
        # deepdiff w/py27 complains about unicode type and val errors
        if 'elem' in dicary:
            dicary['elem'] = [str(e) for e in dicary['elem']]
        if 'elbl' in dicary:
            dicary['elbl'] = [str(l) for l in dicary['elbl']]
        if 'fix_symmetry' in dicary:
            dicary['fix_symmetry'] = str(dicary['fix_symmetry'])
        if 'units' in dicary:
            dicary['units'] = str(dicary['units'])
        if 'fragment_files' in dicary:
            dicary['fragment_files'] = [str(f) for f in dicary['fragment_files']]
        # and about int vs long errors
        if 'molecular_multiplicity' in dicary:
            dicary['molecular_multiplicity'] = int(dicary['molecular_multiplicity'])
        if 'fragment_multiplicities' in dicary:
            dicary['fragment_multiplicities'] = [(m if m is None else int(m))
                                                 for m in dicary['fragment_multiplicities']]
        if 'fragment_separators' in dicary:
            dicary['fragment_separators'] = [(s if s is None else int(s)) for s in dicary['fragment_separators']]
        # forgive generator version changes
        if 'provenance' in dicary:
            dicary['provenance'].pop('version')
        # regularize connectivity ordering
        if 'connectivity' in dicary:
            conn = [(min(at1, at2), max(at1, at2), bo) for (at1, at2, bo) in dicary['connectivity']]
            conn.sort(key=lambda tup: tup[0])
            dicary['connectivity'] = conn

        return dicary

    xptd = massage_dicts(xptd)
    cptd = massage_dicts(cptd)

    if relative_geoms == 'exact':
        pass
    elif relative_geoms == 'align':
        raise FeatureNotImplemented(
            """compare_molrecs(..., relative_geoms='align') not available without B787 from qcdb.""")
        ## can't just expect geometries to match, so we'll align them, check that
        ##   they overlap and that the translation/rotation arrays jibe with
        ##   fix_com/orientation, then attach the oriented geom to computed before the
        ##   recursive dict comparison.
        #from .align import B787
        #cgeom = np.array(cptd['geom']).reshape((-1, 3))
        #rgeom = np.array(xptd['geom']).reshape((-1, 3))
        #rmsd, mill = B787(rgeom=rgeom,
        #                  cgeom=cgeom,
        #                  runiq=None,
        #                  cuniq=None,
        #                  atoms_map=True,
        #                  mols_align=True,
        #                  run_mirror=False,
        #                  verbose=0)
        #if cptd['fix_com']:
        #    compare_integers(1, np.allclose(np.zeros((3)), mill.shift, atol=thresh), 'null shift', verbose=verbose)
        #if cptd['fix_orientation']:
        #    compare_integers(1, np.allclose(np.identity(3), mill.rotation, atol=thresh), 'null rotation', verbose=verbose)
        #ageom = mill.align_coordinates(cgeom)
        #cptd['geom'] = ageom.reshape((-1))

    compare_dicts(xptd, cptd, tol, label, forgive=forgive, verbose=verbose)


#def compare_arrays(expected, computed, digits, label, verbose=1):
#    """Function to compare two numpy arrays. Prints :py:func:`util.success`
#    when elements of vector *computed* match elements of vector *expected* to
#    number of *digits*. Performs a system exit on failure to match symmetry
#    structure, dimension, or element values. Used in input files in the test suite.
#
#    """
#    try:
#        expected = np.asarray(expected)
#        computed = np.asarray(computed)
#        shape1 = expected.shape
#        shape2 = computed.shape
#    except:
#        raise _TestComparisonError("Input objects do not have a shape attribute.")
#
#    if shape1 != shape2:
#        _TestComparisonError("Input shapes do not match.")
#
#    tol = 10 ** (-digits)
#    if not np.allclose(expected, computed, atol=tol):
#        message = "\tArray difference norm is %12.6f." % np.linalg.norm(expected - computed)
#        raise _TestComparisonError(message)
#    if verbose >= 1:
#        _success(label)



import numpy as np
import pytest
from pydantic import ValidationError

from qcelemental.models import (Molecule, ResultInput, Result, OptimizationInput, Optimization, ComputeError,
                                FailedOperation)


@pytest.fixture
def water():
    """Water dimer minima"""
    return Molecule.from_data(
        """
        0 1
        O  -1.551007  -0.114520   0.000000
        H  -1.934259   0.762503   0.000000
        H  -0.599677   0.040712   0.000000
        --
        O   1.350625   0.111469   0.000000
        H   1.680398  -0.373741  -0.758561
        H   1.680398  -0.373741   0.758561
        """,
        dtype="psi4",
        orient=True)


@pytest.fixture
def result_input():
    return {"id": "12345a",
            "driver": "gradient",
            "model": {"method": "filler"}}


@pytest.fixture
def res_success():
    return {"success": True,
            "return_result": 536.2}


@pytest.fixture
def res_failure(res_success):
    return {**res_success,
            "success": False,
            "error": {
                "error_type": "expected_testing_error",
                "error_message": "If you see this, its all good"
            }}


@pytest.fixture
def opti_input():
    return {"input_specification": {
        "driver": "gradient",
        "model": {"method": "filler"}
    }}


@pytest.fixture
def opti_success(water, result_input, res_success):
    res = Result(molecule=water, **result_input, **res_success)
    return {"success": True,
            "trajectory": [res]*3,
            "final_molecule": water,
            "energies": [1.0, 2.0, 3.0]}


@pytest.fixture
def opti_failure(opti_success):
    return {**opti_success,
            "success": False,
            "error": {
                "error_type": "expected_testing_error",
                "error_message": "If you see this, its all good"
            }}


def test_molecule_serialization(water):
    assert isinstance(water.dict(), dict)
    assert isinstance(water.json(), str)


def test_result_pass_serialization(water, result_input, res_success):
    res_in = ResultInput(molecule=water, **result_input)
    assert isinstance(res_in.dict(), dict)
    assert isinstance(res_in.json(), str)
    res_out = Result(molecule=water, **result_input, **res_success)
    assert isinstance(res_out.dict(), dict)
    assert isinstance(res_out.json(), str)


def test_result_wrong_serialization(water, result_input, res_failure):
    res_out = Result(molecule=water, **result_input, **res_failure)
    assert isinstance(res_out.error, ComputeError)
    assert isinstance(res_out.dict(), dict)
    out_json = res_out.json()
    assert isinstance(out_json, str)
    assert 'its all good' in out_json


def test_optimization_pass_serialization(water, opti_input, opti_success):
    opti_in = OptimizationInput(initial_molecule=water, **opti_input)
    assert isinstance(opti_in.dict(), dict)
    assert isinstance(opti_in.json(), str)
    opti_out = Optimization(initial_molecule=water, **opti_input, **opti_success)
    assert isinstance(opti_out.dict(), dict)
    assert isinstance(opti_out.json(), str)


def test_optimization_wrong_serialization(water, opti_input, opti_failure):
    opti_out = Optimization(initial_molecule=water, **opti_input, **opti_failure)
    assert isinstance(opti_out.error, ComputeError)
    assert isinstance(opti_out.dict(), dict)
    out_json = opti_out.json()
    assert isinstance(out_json, str)
    assert 'its all good' in out_json


def test_failed_operation(water, result_input, opti_failure):
    failed = FailedOperation(garbage=water, **result_input, **opti_failure)
    assert isinstance(failed.error, ComputeError)
    assert isinstance(failed.dict(), dict)
    failed_json = failed.json()
    assert isinstance(failed_json, str)
    assert 'its all good' in failed_json



from .from_arrays import from_arrays, from_input_arrays
from .from_string import from_string
from .nucleus import reconcile_nucleus, parse_nucleus_label
from .chgmult import validate_and_fill_chgmult
from .to_string import to_string
from .to_schema import to_schema
from .from_schema import from_schema, contiguize_from_fragment_pattern
